#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['count_99_95_68', 'guess_99_95_68', 'standard_deviation', 'variance', 'statistics', 'average', 'Number']


from typing import Tuple, Sequence, Iterable, Union, Callable
from collections import deque
from math import sqrt
from cengal.math.numbers import *


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def statistics(iterable_readings: Iterable[Number]) -> Tuple[float, Sequence[Number]]:
    """Example
    
        ```
        diff_mean, readings = statistics(iterable_readings)
        ```

    Args:
        iterable_readings (Iterable[Number]): _description_

    Returns:
        Tuple[float, Sequence[Number]]: _description_
    """    
    readings = deque()
    data_sum = 0
    for reading in iterable_readings:
        readings.append(reading)
        data_sum += reading

    try:
        diff_mean = data_sum / len(readings)
    except ZeroDivisionError:
        diff_mean = 0
    
    return diff_mean, readings


def variance(diff_mean: float, readings: Sequence[Number]) -> Tuple[float, float, float, float]:
    """Example:

        ```
        diff_mean, readings = statistics(iterable_readings)
        diff_mean, variance, max_deviation, min_deviation = variance(diff_mean, readings)
        ```

    Alternative example:

        ```
        diff_mean, variance, max_deviation, min_deviation = variance(*statistics(iterable_readings))
        ```

    Args:
        diff_mean (float): _description_
        readings (Sequence[Number]): _description_

    Returns:
        Tuple[float, float, float, float]: _description_
    """    
    max_deviation = None
    min_deviation = None
    deviation_square_sum = 0
    for reading in readings:
        deviation = reading - diff_mean
        deviation_square_sum += deviation ** 2
        if max_deviation is None:
            max_deviation = deviation
        max_deviation = max(max_deviation, deviation)
        if min_deviation is None:
            min_deviation = deviation
        min_deviation = min(min_deviation, deviation)

    try:
        variance = deviation_square_sum / len(readings)
    except ZeroDivisionError:
        variance = 0

    return diff_mean, variance, max_deviation, min_deviation


def standard_deviation(diff_mean, variance, max_deviation, min_deviation) -> Tuple[float, float, float, float]:
    """Example:

        ```
        diff_mean, readings = statistics(iterable_readings)
        diff_mean, variance, max_deviation, min_deviation = variance(diff_mean, readings)
        diff_mean, sd, max_deviation, min_deviation = standard_deviation(diff_mean, variance, max_deviation, min_deviation)
        ```

    Alternative example:

        ```
        diff_mean, sd, max_deviation, min_deviation = standard_deviation(*variance(*statistics(iterable_readings)))
        ```

    Args:
        diff_mean (_type_): _description_
        variance (_type_): _description_
        max_deviation (_type_): _description_
        min_deviation (_type_): _description_

    Returns:
        Tuple[float, float, float, float]: _description_
    """    
    return diff_mean, sqrt(variance), max_deviation, min_deviation


def guess_99_95_68(mean, sd, max_deviation, min_deviation) -> Tuple[float, float, float, float, float]:
    """Example:

        ```
        diff_mean, readings = statistics(iterable_readings)
        diff_mean, variance, max_deviation, min_deviation = variance(diff_mean, readings)
        diff_mean, sd, max_deviation, min_deviation = standard_deviation(diff_mean, variance, max_deviation, min_deviation)
        val_99, val_95, val_68, max_deviation, min_deviation = guess_99_95_68(diff_mean, sd, max_deviation, min_deviation)
        ```

    Alternative example:

        ```
        val_99, val_95, val_68, max_deviation, min_deviation = guess_99_95_68(*standard_deviation(*variance(*statistics(iterable_readings)))
        ```

    Args:
        mean (_type_): _description_
        sd (_type_): _description_
        max_deviation (_type_): _description_
        min_deviation (_type_): _description_

    Returns:
        Tuple[float, float, float, float, float]: _description_
    """    
    # See: https://en.wikipedia.org/wiki/68%E2%80%9395%E2%80%9399.7_rule
    val_68 = mean + sd
    val_95 = mean + 2 * sd
    val_99 = mean + 3 * sd
    return val_99, val_95, val_68, max_deviation, min_deviation


def average(data: Sequence[Number]) -> Number:
    data_len = len(data)
    data_sum = sum(data)
    try:
        return data_sum / data_len
    except ZeroDivisionError:
        return 0


def count_99_95_68(iterable_readings: Iterable[Number], operation: Callable=average) -> Tuple[float, float, float, float, float]:
    """_summary_

    Args:
        iterable_readings (Iterable[Number]): _description_
        operation (Callable, optional): _description_. Defaults to average. You can use other functions like min, max, median, etc.

    Returns:
        Tuple[float, float, float, float, float]: _description_
    """    
    diff_mean, readings = statistics(iterable_readings)
    _, _, max_deviation, min_deviation = variance(diff_mean, readings)
    sorted_readings = sorted(readings)
    readings_qnt = len(sorted_readings)

    def get_slice(data: Sequence, divider: int) -> Sequence:
        data_slice = data[:divider]
        if len(data_slice) == len(data):
            data_slice = data[:len(data)-1]
        return data_slice

    div_68 = round(readings_qnt * 0.68)
    val_68 = operation(get_slice(sorted_readings, div_68))

    div_95 = round(readings_qnt * 0.95)
    val_95 = operation(get_slice(sorted_readings, div_95))

    div_99 = round(readings_qnt * 0.99)
    val_99 = operation(get_slice(sorted_readings, div_99))

    return val_99, val_95, val_68, max_deviation, min_deviation




#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['Instance', 'InstanceRequest', 'fast_wait', 'afast_wait', 'fast_get', 'fast_set',
           'fast_wait_explicit', 'afast_wait_explicit', 'fast_get_explicit', 'fast_set_explicit']

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_tools.await_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.asyncio_loop import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.timer_func_runner import *
from cengal.file_system.file_manager import path_relative_to_current_dir
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.data_manipulation.serialization import *
from typing import Hashable, Tuple, List, Any, Dict, Callable, Type
from cengal.introspection.inspect import get_exception
from cengal.code_flow_control.args_manager import number_of_provided_args
from cengal.math.numbers import RationalNumber


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class InstanceRequest(ServiceRequest):
    def set(self, key: Union[Type, Hashable], instance: Any) -> None:
        return self._save(0, key, instance)
    def get(self, key: Union[Type, Hashable]) -> Any:
        return self._save(1, key)
    def wait(self, key: Union[Type, Hashable]) -> Any:
        return self._save(2, key)


class Instance(DualImmediateProcessingServiceMixin, TypedService[Union[None, Any]], EntityStatsMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(Instance, self).__init__(loop)
        self.instances: Dict = dict()
        self._default_init()
        self.waiters_per_key: Dict[Hashable, Set[CoroID]] = dict()
        self.results: Dict[CoroID, Hashable] = dict()
        self._request_workers = {
            0: self._on_set,
            1: self._on_get,
            2: self._on_wait,
        }

    def get_entity_stats(self, stats_level: 'EntityStatsMixin.StatsLevel' = EntityStatsMixin.StatsLevel.debug) -> Tuple[str, Dict[str, Any]]:
        return type(self).__name__, {
            'instances num': len(self.instances),
        }
    
    def _default_init(self):
        from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import CoroPriority, GlyPatchManager
        from cengal.code_flow_control.smart_values import ValueHolder
        from cengal.parallel_execution.asyncio.efficient_streams import StreamManagerIOCoreMemoryManagement, TcpStreamManager, UdpStreamManager
        from cengal.io.core.memory_management import IOCoreMemoryManagement
        from cengal.io.named_connections.named_connections_manager import NamedConnectionsManager
        from cengal.data_manipulation.conversion.reinterpret_cast_management.standard_library.copy_wrapper import CopyWrapper
        from cengal.data_manipulation.conversion.reinterpret_cast_management.standard_library.deep_copy_wrapper import DeepCopyWrapper
        from cengal.data_manipulation.conversion.reinterpret_cast_management.standard_library.uni_copy_wrapper import UniCopyWrapper
        from cengal.file_system.app_fs_structure.app_dir_path import AppDirPath, AppDirectoryType
        
        data_full_size: ValueHolder[RationalNumber] = ValueHolder(True, 0)
        other_data_full_size: ValueHolder[RationalNumber] = ValueHolder(True, 0)
        in_data_full_size: ValueHolder[RationalNumber] = ValueHolder(True, 0)
        out_data_full_size: ValueHolder[RationalNumber] = ValueHolder(True, 0)
        tcp_stream_manager: TcpStreamManager = TcpStreamManager()
        tcp_stream_manager.io_memory_management.global__data_full_size = data_full_size
        tcp_stream_manager.io_memory_management.global_other__data_full_size = other_data_full_size
        tcp_stream_manager.io_memory_management.global_in__data_full_size = in_data_full_size
        tcp_stream_manager.io_memory_management.global_out__data_full_size = out_data_full_size
        udp_stream_manager: UdpStreamManager = UdpStreamManager()
        udp_stream_manager.io_memory_management.global__data_full_size = data_full_size
        udp_stream_manager.io_memory_management.global_other__data_full_size = other_data_full_size
        udp_stream_manager.io_memory_management.global_in__data_full_size = in_data_full_size
        udp_stream_manager.io_memory_management.global_out__data_full_size = out_data_full_size
        named_connections_manager: NamedConnectionsManager = NamedConnectionsManager(data_full_size)
        copy_wrapper = CopyWrapper()
        deep_copy_wrapper = DeepCopyWrapper()
        uni_copy_wrapper = UniCopyWrapper()
        app_dir_path = AppDirPath()
        gly_patch_manager = GlyPatchManager()

        self.instances.update({
            'data_full_size': data_full_size,
            'other_data_full_size': other_data_full_size,
            'in_data_full_size': in_data_full_size,
            'out_data_full_size': out_data_full_size,
            TcpStreamManager: tcp_stream_manager,
            UdpStreamManager: udp_stream_manager,
            NamedConnectionsManager: named_connections_manager,
            CopyWrapper: copy_wrapper,
            DeepCopyWrapper: deep_copy_wrapper,
            UniCopyWrapper: uni_copy_wrapper,
            'app_name': str(),
            'app_name_for_fs': str(),
            'app_version': tuple(),
            'app_version_str': str(),
            AppDirPath: app_dir_path,
            'app_data_dir_path_type': AppDirectoryType.local_data,
            'app_cache_dir_path_type': AppDirectoryType.local_cache,
            'app_temp_dir_path_type': AppDirectoryType.local_temp,
            'app_log_dir_path_type': AppDirectoryType.local_log,
            'app_config_dir_path_type': AppDirectoryType.local_config,
            'app_runtime_dir_path_type': AppDirectoryType.local_runtime,
            GlyPatchManager: gly_patch_manager,
        })

    def single_task_registration_or_immediate_processing_single(
            self, *args, **kwargs
    ) -> Tuple[bool, Union[Type, Hashable], Optional[BaseException]]:
        number_of_args: int = number_of_provided_args(args, kwargs)
        if 1 == number_of_args:
            return self._on_get(*args, **kwargs)
        elif 2 == number_of_args:
            return self._on_set(*args, **kwargs)
        else:
            return True, None, RuntimeError(f'Wrong number of parameters: {number_of_args}')

    def full_processing_iteration(self):
        results_bak = self.results
        self.results = type(results_bak)()
        for waiter_coro_id, instance in results_bak.items():
            self.register_response(waiter_coro_id, instance)

        self.make_dead()

    def in_work(self) -> bool:
        return self.thrifty_in_work(bool(self.results))
    
    def _on_set(self, key: Union[Type, Hashable], instance: Any):
        self.instances[key] = instance
        if key in self.waiters_per_key:
            self.make_live()
            waiters = self.waiters_per_key[key]
            del self.waiters_per_key[key]
            for waiter_coro_id in waiters:
                self.results[waiter_coro_id] = instance
        
        return True, None, None
    
    def _on_get(self, key: Union[Type, Hashable]):
        instance = None
        exception = None
        try:
            instance = self.instances[key]
        except:
            exception = get_exception()
        
        return True, instance, exception
    
    def _on_wait(self, key: Union[Type, Hashable]):
        instance = None
        exception = None
        need_to_wait: bool = False
        try:
            instance = self.instances[key]
        except KeyError:
            need_to_wait = True
        except:
            exception = get_exception()
        
        if need_to_wait:
            current_coro_id = self.current_caller_coro_info.coro_id
            if key not in self.waiters_per_key:
                self.waiters_per_key[key] = set()
            
            self.waiters_per_key[key].add(current_coro_id)
            return False, None, None
        else:
            return True, instance, exception
    
    def inline_set(self, key: Union[Type, Hashable], instance: Any):
        self.instances[key] = instance
    
    def inline_get(self, key: Union[Type, Hashable]) -> Any:
        return self.instances[key]


InstanceRequest.default_service_type = Instance


def fast_wait_explicit(i: Interface, key: Union[Type, Hashable]) -> Any:
    loop: CoroSchedulerType = i._loop
    instance_service: Instance = loop.get_service_instance_fast(Instance)
    need_to_wait: bool = False
    try:
        return instance_service.inline_get(key)
    except KeyError:
        need_to_wait = True
    
    if need_to_wait:
        return i(InstanceRequest().wait(key))


def fast_wait(key: Union[Type, Hashable]) -> Any:
    i: Interface = current_interface()
    loop: CoroSchedulerType = i._loop
    instance_service: Instance = loop.get_service_instance_fast(Instance)
    need_to_wait: bool = False
    try:
        return instance_service.inline_get(key)
    except KeyError:
        need_to_wait = True
    
    if need_to_wait:
        return i(InstanceRequest().wait(key))


async def afast_wait_explicit(i: Interface, key: Union[Type, Hashable]) -> Any:
    loop: CoroSchedulerType = i._loop
    instance_service: Instance = loop.get_service_instance_fast(Instance)
    need_to_wait: bool = False
    try:
        return instance_service.inline_get(key)
    except KeyError:
        need_to_wait = True
    
    if need_to_wait:
        return await i(InstanceRequest().wait(key))


async def afast_wait(key: Union[Type, Hashable]) -> Any:
    i: Interface = current_interface()
    loop: CoroSchedulerType = i._loop
    instance_service: Instance = loop.get_service_instance_fast(Instance)
    need_to_wait: bool = False
    try:
        return instance_service.inline_get(key)
    except KeyError:
        need_to_wait = True
    
    if need_to_wait:
        return await i(InstanceRequest().wait(key))


def fast_get_explicit(i: Interface, key: Union[Type, Hashable]) -> Any:
    loop: CoroSchedulerType = i._loop
    instance_service: Instance = loop.get_service_instance_fast(Instance)
    return instance_service.inline_get(key)


def fast_get(key: Union[Type, Hashable]) -> Any:
    i: Interface = current_interface()
    loop: CoroSchedulerType = i._loop
    instance_service: Instance = loop.get_service_instance_fast(Instance)
    return instance_service.inline_get(key)


def fast_set_explicit(i: Interface, key: Union[Type, Hashable], instance: Any):
    loop: CoroSchedulerType = i._loop
    instance_service: Instance = loop.get_service_instance_fast(Instance)
    instance_service.inline_set(key, instance)


def fast_set(key: Union[Type, Hashable], instance: Any):
    i: Interface = current_interface()
    loop: CoroSchedulerType = i._loop
    instance_service: Instance = loop.get_service_instance_fast(Instance)
    instance_service.inline_set(key, instance)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import threading
from contextlib import contextmanager


"""
This module is usefull for Linux kernels before 3.17 (Ubuntu before 15.04). It is not usefull for Windows.
Before version 3.17, fetching the current time in the Linux kernel involved a system call, which is relatively expensive in terms of performance.
It was Ubuntu 15.04 (Vivid Vervet) that included Linux kernel 3.19 by default, so it benefited from the improved time fetching mechanism introduced in 3.17
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__server_clock_var = 0
__global_timer = None
__global_semaphore = threading.RLock()
__timer_interval = 0.2  # in seconds (float value)
__is_need_to_stop_clock = False


def get_server_clock():
    # return time.time()

    global __server_clock_var
    global __global_timer
    global __global_semaphore
    global __is_need_to_stop_clock

    if __global_timer is None:
        with __global_semaphore:
            __is_need_to_stop_clock = False
            __server_clock_var = time.time()
            __global_timer = threading.Timer(__timer_interval, __update_server_clock)
            __global_timer.start()
    result = __server_clock_var

    return result


def stop_server_clock():
    global __server_clock_var
    global __global_timer
    global __global_semaphore
    global __is_need_to_stop_clock

    with __global_semaphore:
        __is_need_to_stop_clock = True
        if __global_timer is not None:
            __global_timer.cancel()


def __update_server_clock():
    global __server_clock_var
    global __global_timer
    global __global_semaphore
    global __is_need_to_stop_clock

    with __global_semaphore:
        __server_clock_var = time.time()
        if not __is_need_to_stop_clock:
            __global_timer = threading.Timer(__timer_interval, __update_server_clock)
            __global_timer.start()


@contextmanager
def server_clock():
    try:
        yield get_server_clock
    except:
        raise
    finally:
        stop_server_clock()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.introspection.inspect import get_exception
from typing import Callable, Any, Optional, Tuple, List, Dict, Type
from enum import Enum
import queue
import threading
import sys


ExceptionInfo = Tuple


class RequestToThread:
    class Command(Enum):
        request = 0
        shut_down = 1

    def __init__(self, command_type: 'Command', request: Any):
        self.command_type = command_type
        self.request = request


class ResponseFromThread:
    def __init__(self, has_response: bool, response: Any, exception: Optional[ExceptionInfo] = None):
        self.has_response = has_response  # type: bool
        self.response = response          # type: Any
        self.exception = exception        # type: Optional[ExceptionInfo]


ThreadWorker = Callable[[RequestToThread], ResponseFromThread]
TypeOfThreadWorker = Type[ThreadWorker]


class ServiceThread(threading.Thread):
    def __init__(self, worker_type: TypeOfThreadWorker):
        super(ServiceThread, self).__init__(daemon=True)
        self.requests = queue.Queue()  # type: queue.Queue
        self.results = queue.Queue()   # type: queue.Queue
        self.worker = worker_type()    # type: ThreadWorker

    def run(self):
        shut_down = False
        while not shut_down:
            request: RequestToThread = self.requests.get()

            if RequestToThread.Command.shut_down == request.command_type:
                shut_down = True

            try:
                response = self.worker(request)
            except:
                response = ResponseFromThread(False, None, get_exception())

            self.results.put(response)
            self.requests.task_done()

    def put(self, request: RequestToThread):
        self.requests.put(request)

    def put_nowait(self, request: RequestToThread):
        self.requests.put_nowait(request)

    def get(self) -> Any:
        response: ResponseFromThread = self.results.get()
        self.results.task_done()
        if response.has_response:
            return response.response
        elif response.exception is not None:
            raise response.exception[1]
        else:
            raise RuntimeError

    def get_nowait(self) -> Any:
        response: ResponseFromThread = self.results.get_nowait()
        self.results.task_done()
        if response.has_response:
            return response.response
        elif response.exception is not None:
            raise response.exception[1]
        else:
            raise RuntimeError


class ServiceThreadPool:
    def __init__(self, worker_type: TypeOfThreadWorker, number_of_threads: int = 1):
        self.worker_type = worker_type
        self.number_of_threads = number_of_threads
        if self.number_of_threads < 1:
            self.number_of_threads = 1
        self.threads = list()                 # type: List[ServiceThread]
        self.thread_pending_results = dict()  # type: Dict[ServiceThread, int]
        self.pending_requests_queue = list()  # type: List[RequestToThread]
        self._init()

    def put_synchronous(self, request: RequestToThread):
        thread: ServiceThread = self._get_best_thread()
        thread.put(request)
        self.thread_pending_results[thread] += 1

    def put_into_pending_queue(self, request: RequestToThread):
        self.pending_requests_queue.append(request)

    def put_pending_queue_into_work(self):
        buff_pending_requests_queue = self.pending_requests_queue
        self.pending_requests_queue = type(self.pending_requests_queue)()
        for pending_request in buff_pending_requests_queue:
            if not self._put_impl(pending_request):
                self.pending_requests_queue.append(pending_request)

    def get_results(self) -> List[Any]:
        responses: List[Any] = list()
        for thread, pending_responses in self.thread_pending_results.items():
            if pending_responses:
                try:
                    while True:
                        responses.append(thread.get_nowait())
                        self.thread_pending_results[thread] -= 1
                except queue.Empty:
                    pass
        return responses

    def stop(self) -> List[Any]:
        for thread in self.threads:
            thread.put(RequestToThread(RequestToThread.Command.shut_down, None))
        for thread in self.threads:
            thread.join()
        return self.get_results()

    def _init(self):
        for index in range(self.number_of_threads):
            self.threads.append(ServiceThread(self.worker_type))
        for thread in self.threads:
            self.thread_pending_results[thread] = 0
            thread.start()

    def _get_best_thread(self) -> ServiceThread:
        thread_load: Dict = dict()
        thread_index = -1
        for thread in self.threads:
            thread: ServiceThread = thread
            thread_index += 1
            thread_load[thread_index] = thread.requests.unfinished_tasks
        sorted_by_value = sorted(thread_load.items(), key=lambda kv: kv[1])
        return self.threads[sorted_by_value[0][0]]

    def _get_threads_list(self) -> List[Tuple[int, ServiceThread]]:
        thread_load: Dict = dict()
        thread_index = -1
        for thread in self.threads:
            thread: ServiceThread = thread
            thread_index += 1
            thread_load[thread_index] = thread.requests.unfinished_tasks
        sorted_by_value = sorted(thread_load.items(), key=lambda kv: kv[1])
        return sorted_by_value

    def _put_impl(self, request: RequestToThread) -> bool:
        is_ok = False
        for thread_index, thread_qsize in self._get_threads_list():
            try:
                thread = self.threads[thread_index]
                thread.put_nowait(request)
                self.thread_pending_results[thread] += 1
                is_ok = True
                break
            except queue.Full:
                pass
        return is_ok



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['AutoDerivedClass', 'BaseAutoDerivedObjWrapper', 'ClassWrappingFactory', 'ClassWrappingFactoryWithObjDataSizeLog']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from sys import getsizeof
from cengal.data_generation.id_generator import IDGenerator
from cengal.data_manipulation.conversion.reinterpret_cast import TemporaryReinterpretCast, reinterpret_cast
from cengal.entities.copyable import copy__impl, deepcopy__impl
from typing import Dict, Type, Callable, Any, Tuple, FrozenSet, Set, Optional
from copy import Error as CopyError, copy, deepcopy
from socket import SocketIO


class AutoDerivedClass:
    instance_id = IDGenerator()

    def __init__(self):
        self._instance_id: int = self.instance_id()
        self.class_name = type(self).__name__
        self.derived: Dict[Type, Type] = dict()
    
    @property
    def methods(self) -> Dict[str, Callable]:
        raise NotImplementedError
    
    def base_classes(self, base_type) -> Tuple[Type]:
        return (base_type,)
    
    def __call__(self, base_type: Type):
        base_type_name = base_type.__name__
        try:
            return self.derived[base_type_name]
        except KeyError:
            result: Type = type(f'{self.class_name}__{self._instance_id}__from__{base_type_name}', self.base_classes(base_type), self.methods)
            self.derived[base_type_name] = result
            return result


class BaseAutoDerivedObjWrapper:
    instance_id = IDGenerator()

    def __init__(self):
        self._instance_id: int = self.instance_id()
        self.class_name = type(self).__name__
        self.derived: Dict[Type, Type] = dict()
        self.t = self.type
    
    def wrapping_required(self, obj: Any, base_type: Type, fields: Tuple[str], planned_type_name: str) -> bool:
        return True
    
    def methods(self, obj: Any, base_type: Type, fields: Tuple[str]) -> Dict[str, Callable]:
        raise NotImplementedError

    def base_classes(self, obj: Any, base_type: Type, fields: Tuple[str]) -> Tuple[Type]:
        return (base_type,)
    
    def gen_fields_tuple(self, obj: Any):
        try:
            return obj.__slots__
        except AttributeError:
            pass

        return tuple(obj.__dict__.keys())
    
    def type(self, obj: Any):
        base_type = type(obj)
        base_type_name = base_type.__name__
        try:
            return self.derived[base_type_name]
        except KeyError:
            fields: Tuple = self.gen_fields_tuple(obj)
            planned_type_name: str = f'{self.class_name}__{self._instance_id}__from__{base_type_name}'
            if self.wrapping_required(obj, base_type, fields, planned_type_name):
                result: Type = type(planned_type_name, self.base_classes(obj, base_type, fields), self.methods(obj, base_type, fields))
                self.derived[base_type_name] = result
                return result
            else:
                return base_type
    
    def __call__(self, obj: Any):
        obj.__class__ = self.type(obj)
        return obj

    def temporary(self, obj: Any):
        return TemporaryReinterpretCast(self.type(obj), obj)

    def persistent(self, obj: Any):
        return reinterpret_cast(self.type(obj), obj)


# class ClassWrapper:
#     def __init__(self, auto_derived_types: AutoDerivedTyped) -> None:
#         self._class_wrapper__data_size: int = 0
#         self._class_wrapper__auto_derived_types: AutoDerivedTyped = auto_derived_types
    
#     def __call__(self, *args: Any, **kwds: Any) -> Any:
#         pass


class ClassWrappingFactory:
    instance_id = IDGenerator()

    def __init__(self):
        self._instance_id: int = self.instance_id()
        self.class_name = type(self).__name__
        self.derived: Dict[Type, Type] = dict()
        self._attributes_and_methods: Dict[str, Callable] = {
            '_class_wrapper__field_names': None,
            '_class_wrapper__factory': self,
        }
    
    @property
    def attributes_and_methods(self) -> Dict[str, Callable]:
        return self._attributes_and_methods.copy()
    
    def edit_derived_type(self, obj: Any, attributes_and_methods, field_names: Tuple[str]) -> Dict[str, Any]:
        """Might be overloaded in a derived class if needed

        Args:
            obj (Any): _description_
            attributes_and_methods (_type_): _description_
            field_names (Tuple[str]): _description_
            data_size (int): _description_

        Returns:
            Dict[str, Any]: _description_
        """        
        return attributes_and_methods
    
    def gen_kwargs(self, obj: Any, attributes_and_methods, field_names: Tuple[str]) -> Optional[Dict[str, Any]]:
        """Might be overloaded in a derived class if needed

        Args:
            obj (Any): _description_
            attributes_and_methods (_type_): _description_
            field_names (Tuple[str]): _description_
            data_size (int): _description_

        Returns:
            Optional[Dict[str, Any]]: _description_
        """        
        return None
    
    def __call__(self, obj: Any):
        base_type = type(obj)
        base_type_name = base_type.__name__
        try:
            return self.derived[base_type_name]
        except KeyError:
            field_names: Tuple = self.gen_fields_tuple(obj)
            
            attributes_and_methods = self.attributes_and_methods
            attributes_and_methods['_class_wrapper__field_names'] = field_names
            attributes_and_methods = self.edit_derived_type(obj, attributes_and_methods, field_names)
            kwargs = self.gen_kwargs(obj, attributes_and_methods, field_names)
            
            if kwargs is None:
                derived: Type = type(f'{self.class_name}__{self._instance_id}__from__{base_type_name}', (base_type,), attributes_and_methods)
            else:
                derived: Type = type(f'{self.class_name}__{self._instance_id}__from__{base_type_name}', (base_type,), attributes_and_methods, **kwargs)

            self.derived[base_type_name] = derived
            return derived
    
    def gen_fields_tuple(self, obj: Any):
        try:
            return obj.__slots__
        except AttributeError:
            pass

        try:
            return tuple(obj.__dict__.keys())
        except AttributeError:
            pass
        
        return tuple()


class ClassWrappingFactoryWithObjDataSizeLog(ClassWrappingFactory):
    def __init__(self):
        super().__init__()
        self.attributes_and_methods['_class_wrapper__obj_data_size_log'] = 0
    
    def edit_derived_type(self, obj: Any, attributes_and_methods, field_names: Tuple[str], data_size: int) -> Dict[str, Any]:
        data_size = 0
        for field_name in field_names:
            data_size += getsizeof(getattr(obj, field_name))

        attributes_and_methods['_class_wrapper__obj_data_size_log'] = data_size
        return attributes_and_methods



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['cvalue_to_dict']


from ctypes import Structure, _Pointer, Array, c_char, c_wchar
from typing import Dict


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def cvalue_to_dict(item) -> Dict:
    """_summary_
        Example:
            from pprintpp import pprint
            
            def wnd_proc_impl(self, hwnd: wintypes.HWND, msg: wintypes.UINT, wparam: wintypes.WPARAM, lparam: wintypes.LPARAM) -> LRESULT:
                if win32con.WM_NCCALCSIZE == msg:
                    lpnccalcsize_params = ctypes.cast(lparam, LPNCCALCSIZE_PARAMS)
                    print(f'WM_NCCALCSIZE: lpnccalcsize_params=')
                    pprint(ctypes_to_dict(lpnccalcsize_params))
                    print()
        
        Output:
            WM_NCCALCSIZE: lpnccalcsize_params=
            {
                'LP_NCCALCSIZE_PARAMS': {
                    'contents': {
                        'NCCALCSIZE_PARAMS': {
                            'lppos': {
                                'LP_WINDOWPOS': {
                                    'contents': {
                                        'WINDOWPOS': {
                                            'cx': 416,
                                            'cy': 240,
                                            'flags': 6199,
                                            'hwnd': 22679738,
                                            'hwndInsertAfter': None,
                                            'x': 507,
                                            'y': 114,
                                        },
                                    },
                                },
                            },
                            'rgrc': {
                                'RECT_Array_3': [
                                    {
                                        'RECT': {
                                            'bottom': 354,
                                            'left': 507,
                                            'right': 923,
                                            'top': 114,
                                        },
                                    },
                                    {
                                        'RECT': {
                                            'bottom': 354,
                                            'left': 507,
                                            'right': 923,
                                            'top': 114,
                                        },
                                    },
                                    {
                                        'RECT': {
                                            'bottom': 354,
                                            'left': 507,
                                            'right': 923,
                                            'top': 114,
                                        },
                                    },
                                ],
                            },
                        },
                    },
                },
            }
    
    Args:
        item (_type_): _description_

    Returns:
        Dict: _description_
    """    
    item_type_name = type(item).__name__
    if isinstance(item, Structure):
        st = item
        data = dict()
        fields = st._fields_
        for field_name, field_type in fields:
            field_value = cvalue_to_dict(getattr(st, field_name))
            data[field_name] = field_value
        
        return {item_type_name: data}
    elif isinstance(item, _Pointer):
        field_name = 'contents'
        field_value = cvalue_to_dict(item.contents)
        return {item_type_name: {field_name: field_value}}
    elif isinstance(item, Array):
        array_type = item._type_
        if c_char == array_type:
            return item.value  # bytes
        elif c_wchar == array_type:
            return item.value  # str
        else:
            data = list()
            array_len = len(item)
            for sub_item in item:
                data.append(cvalue_to_dict(sub_item))

            return {item_type_name: data}
    else:
        return item




#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['tag_db_interface', 'SMART_TREE_TYPE', 'SMART_TREE_TYPE_WITH_INTERNAL_MENU', 'FULL_TREE_TYPE', 'PLAIN_PSEUDO_TREE_TYPE', 'USUAL_TREE_TYPE', 'ToManyIdenticalItemsOnTheGivenTagPathError', 'UnknownTreeTypeError', ]
__author__ = 'Mikhail Butenko <gtalk@mikhail-butenko.in.ua>'


# import TagDB
# from TagDB import USUAL_TREE_TYPE as USUAL_TREE_TYPE
# from TagDB import SMART_TREE_TYPE as SMART_TREE_TYPE
# from TagDB import FULL_TREE_TYPE as FULL_TREE_TYPE
# from TagDB import PLAIN_PSEUDO_TREE_TYPE as PLAIN_PSEUDO_TREE_TYPE
from .TagDB import *
import copy
from cengal.web_tools.request_cache import RequestCache
import time
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import gly, CoroPriority
from cengal.parallel_execution.coroutines.coro_standard_services.read_write_locker import grwl, RWOperation
from cengal.parallel_execution.coroutines.coro_scheduler import EntityStatsMixin
from cengal.data_manipulation.tree_traversal import *


class tag_db_interface(EntityStatsMixin):
    def __init__(self, max_writers_in_progress: int, max_readers_in_progress: int, lock_entity_id: Hashable = None, default_priority: CoroPriority = CoroPriority.normal):
        self.default_priority: CoroPriority = default_priority
        self.__db = TagDB(default_priority)
        self.cache = RequestCache(300000, default_priority=default_priority)
        self.lock_entity_id: Hashable = lock_entity_id or type(self)
        self.max_writers_in_progress: int = max_writers_in_progress
        self.max_readers_in_progress: int = max_readers_in_progress
        self.last_fill_cache_result_time: float = None
        self.last_fill_cache_tags = None
        self.last_tested_fill_cache_result_time: float = None
        self.last_tested_fill_cache_tags = None
        self.fill_cache_iterations: int = 0
        self.fill_cache_iterations_per_second: float = None
        self.fill_cache_last_time_measurement: float = None
        self.fill_cache_last_time_measurement_iterations: int = 0

    def get_entity_stats(self, stats_level: 'EntityStatsMixin.StatsLevel' = EntityStatsMixin.StatsLevel.debug) -> Tuple[str, Dict[str, Any]]:
        tag_db_name, tag_db_stats = self.__db.get_entity_stats(stats_level)
        return type(self).__name__, {
            'interface': {
                'cache size': len(self.cache._requestsAndData),
                'last fill cache result time': self.last_fill_cache_result_time,
                'last fill cache tags': self.last_fill_cache_tags,
                'last tested fill cache result time': self.last_tested_fill_cache_result_time,
                'last tested fill cache tags': self.last_tested_fill_cache_tags,
                'fill cache iterations': self.fill_cache_iterations,
                'fill cache iterations per second': self.fill_cache_iterations_per_second,
            },
            'db': {
                tag_db_name: tag_db_stats,
            }
        }
    
    def _glock(self):
        return grwl(self.lock_entity_id, self.max_writers_in_progress, self.max_readers_in_progress, False)
    
    def _wlock(self):
        return self._glock()(RWOperation.write)
    
    def _rlock(self):
        return self._glock()(RWOperation.read)

    def get_db(self):
        return self.__db
    
    def new_empty(self) -> 'tag_db_interface':
        return tag_db_interface(
            self.max_writers_in_progress,
            self.max_readers_in_progress,
            self.lock_entity_id,
            self.default_priority
            )

    def add_item(self, binItem, binTags):
        with self._wlock():
            binItem = copy.deepcopy(binItem)
            binTags = copy.deepcopy(binTags)

            result = self.__db.add_item(binItem, binTags)

            result = copy.deepcopy(result)
            self.cache.clear()
            return result

    def remove_item(self, binTags, binItem):
        with self._wlock():
            binTags = copy.deepcopy(binTags)
            binItem = copy.deepcopy(binItem)

            result = self.__db.remove_item(binTags, binItem)

            result = copy.deepcopy(result)
            self.cache.clear()
            return result

    def get_items(self, binTags, treeType=USUAL_TREE_TYPE):
        with self._rlock():
            binTags = copy.deepcopy(binTags)
            treeType = copy.deepcopy(treeType)

            # request = ('get_items', tuple(self.__db.sort_tag_list_by_hash(binTags)), treeType)
            request = ('get_items', frozenset(binTags), treeType)
            result = self.cache.try_to_get_data_for_request(request)
            if result is None:
                
                self.cache.put_new_request(request, result)

            result = copy.deepcopy(result)
            return result

    def get_tags(self, binTags, treeType=USUAL_TREE_TYPE):
        with self._rlock():
            binTags = copy.deepcopy(binTags)
            treeType = copy.deepcopy(treeType)

            # request = ('get_tags', tuple(self.__db.sort_tag_list_by_hash(binTags)), treeType)
            request = ('get_tags', frozenset(binTags), treeType)
            result = self.cache.try_to_get_data_for_request(request)
            if result is None:
                result = self.__db.get_tags_from_tags(binTags, treeType=treeType)
                self.cache.put_new_request(request, result)

            result = copy.deepcopy(result)
            return result

    def get_all(self, binTags, treeType=USUAL_TREE_TYPE):
        with self._rlock():
            binTags = copy.deepcopy(binTags)
            treeType = copy.deepcopy(treeType)

            # request = ('get_all', tuple(self.__db.sort_tag_list_by_hash(binTags)), treeType)
            request = ('get_all', frozenset(binTags), treeType)
            result = self.cache.try_to_get_data_for_request(request)
            if result is None:
                result = self.__db.get_all_from_tags(binTags, treeType=treeType)
                self.cache.put_new_request(request, result)

            result = copy.deepcopy(result)
            return result

    def get_tags_for_a_smart_redirection(self, binTags):
        with self._rlock():
            binTags = copy.deepcopy(binTags)

            # request = ('is_smart_redirection_for_a_tag_path_reduction_needed'
            #            , tuple(self.__db.sort_tag_list_by_hash(binTags)))
            request = ('is_smart_redirection_for_a_tag_path_reduction_needed', frozenset(binTags))
            result = self.cache.try_to_get_data_for_request(request)
            if result is None:
                self.__db.get_tags_for_a_smart_redirection(binTags)
                self.cache.put_new_request(request, result)

            result = copy.deepcopy(result)
            return result

    def fill_cache(self, max_answer_time_in_seconds=0.5, path=list()):
        ly = gly(self.default_priority)
        
        current_path = list(path)
        startTime = time.perf_counter()
        self.last_tested_fill_cache_tags = current_path

        tags_and_items_inside_curent_path = self.get_all(current_path, SMART_TREE_TYPE)
        internal_tags = tags_and_items_inside_curent_path[0]
        self.get_items(current_path, PLAIN_PSEUDO_TREE_TYPE)
        
        self.fill_cache_iterations += 1

        endTime = time.perf_counter()
        resultTime = endTime - startTime
        self.last_tested_fill_cache_result_time = resultTime
        if resultTime >= max_answer_time_in_seconds:
            self.last_fill_cache_tags = current_path
            self.last_fill_cache_result_time = resultTime
            for additional_tag in internal_tags:
                ly()
                new_path = current_path + [additional_tag]
                self.fill_cache(max_answer_time_in_seconds, new_path)

    def fill_web_cache(self, depth: int, max_answer_time_in_seconds=0.5, path=None):
        ly = gly(self.default_priority)
        ly()

        path = path or set()
        current_path = list(path)
        startTime = time.perf_counter()

        result, is_cached = self.web_browse_to(current_path, True)
        optimized_path, tags, items = result
        current_path = list(optimized_path)
        self.last_tested_fill_cache_tags = current_path
        
        self.fill_cache_iterations += 1
        depth -= 1
        
        if self.fill_cache_last_time_measurement_iterations == 0:
            self.fill_cache_last_time_measurement_iterations = self.fill_cache_iterations
        
        if self.fill_cache_last_time_measurement is None:
            self.fill_cache_last_time_measurement = time.perf_counter()
        else:
            dtime = time.perf_counter() - self.fill_cache_last_time_measurement
            if dtime >= 1.0:
                self.fill_cache_last_time_measurement = time.perf_counter()
                iterations = self.fill_cache_iterations - self.fill_cache_last_time_measurement_iterations
                self.fill_cache_iterations_per_second = iterations / dtime
                self.fill_cache_last_time_measurement_iterations = self.fill_cache_iterations

        endTime = time.perf_counter()
        resultTime = endTime - startTime
        self.last_tested_fill_cache_result_time = resultTime
        # if (depth > 0) and (is_cached or (resultTime >= max_answer_time_in_seconds)):
        # if (depth > 0) and (resultTime >= max_answer_time_in_seconds):
        if depth > 0:
            self.last_fill_cache_tags = current_path
            self.last_fill_cache_result_time = resultTime
            for additional_tag in tags:
                new_path = current_path + [additional_tag]
                self.fill_web_cache(depth, max_answer_time_in_seconds, new_path)
    
    def clear_cache(self):
        self.cache.clear()
        self.fill_cache_iterations = 0
    
    def web_browse_to(self, binTags, return_is_cached_result: bool = False, use_cache: bool = True):
        with self._rlock():
            binTags = copy.deepcopy(binTags)
            additional_tags = self.get_tags_for_a_smart_redirection(binTags)
            if additional_tags is not None:
                optimized_path = set(binTags) | set(additional_tags)
            else:
                optimized_path = binTags

            request = ('web_browse_to', frozenset(optimized_path))
            result = None
            if use_cache:
                result = self.cache.try_to_get_data_for_request(request)
                is_cached_result = True
            
            if result is None:
                is_cached_result = False
                items, items_ids = self.__db.get_items_from_tags(optimized_path, treeType=PLAIN_PSEUDO_TREE_TYPE, isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags=True)
                tags = self.__db.get_tags_from_tags(optimized_path, treeType=SMART_TREE_TYPE, prePreparedSetOfAllInternalItemIDsForThisSetOfTags=items_ids)
                result = (optimized_path, tags, items)
                if use_cache:
                    self.cache.put_new_request(request, result)

            result = copy.deepcopy(result)
            if return_is_cached_result:
                return result, is_cached_result
            else:
                return result
    
    def fs_browse_to(self, binTags):
        with self._rlock():
            binTags = copy.deepcopy(binTags)

            request = ('web_browse_to', frozenset(binTags))
            result = self.cache.try_to_get_data_for_request(request)
            if result is None:
                optimized_path = set(binTags) | set(self.__db.get_tags_for_a_smart_redirection(binTags))
                items = self.__db.get_items_from_tags(optimized_path, treeType=SMART_TREE_TYPE)
                tags = self.__db.get_tags_from_tags(optimized_path, treeType=SMART_TREE_TYPE)
                result = (optimized_path, tags, items)
                self.cache.put_new_request(request, result)

            result = copy.deepcopy(result)
            return result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython
from enum import Enum
from libc.time cimport clock, CLOCKS_PER_SEC, clock_t
perf_counter = process_time = clock

try:
    from cengal.time_management.cpu_clock_cycles import perf_counter
except ImportError:
    try:
        from time import perf_counter
    except ImportError:
        pass

try:
    from time import process_time
except ImportError:
    pass

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


if cython.compiled:
    # print('"{}" - compiled.'.format(__name__))
    pass
else:
    print('"{}" - interpreted script.'.format(__name__))


class TimeLimitIsTooSmall(Exception):
    def __init__(self, min_time, *args):
        super().__init__(*args)
        self.min_time = min_time


# TODO: add support for cpu_ticks_count
class ClockType(Enum):
    fake = 0
    clock = 1
    perf_counter = 2
    process_time = 3


cdef inline clock_t _fake():
    return 0


cdef inline clock_t _cpython__clock():
    return clock()


cdef c_clock():
    return <float>_cpython__clock() / <float>CLOCKS_PER_SEC


cdef inline clock_t _perf_counter():
    return <clock_t>round(perf_counter() * <float>CLOCKS_PER_SEC)


cdef inline clock_t _process_time():
    return <clock_t>round(process_time() * <float>CLOCKS_PER_SEC)


cdef class BaseTracer:
    """
    Base class of all tracers

    Lets assume you have a tracer:

        tr = Tracer(10.0)

    Or a fake tracer:

        tr = TracerCounter(10000, 10.0)

    As result you can use next interfaces

        tr()  # Will return True if tracer has finished it's counting and as a result, the specified time was passed.

        tr.iter_per_time_unit  # Will return counted time of iterations per second (if time() function uses second
            as a time unit)

        tr.iterations_made  # Will return number of all iterations made (including those that were produced after the
            end of the counting)

        tr.time_spent  # Will return time spent (including time that were used after the end of the counting)

        tr.clock_type  # Will return used time function type (ClockType enum)

    """

    cdef:
        clock_t _cpython__run_time
        unsigned int _cpython__number_of_iterations
        unsigned int _cpython__last_tracked_number_of_iterations
        clock_t _cpython__start_time
        clock_t _cpython__end_time
        clock_t _cpython__relevant_stop_time
        bint _cpython__last_run_was_made
        object _iter_impl
        object _cpython__clock
        object _cpython__clock_type

    def __init__(self, float run_time, object clock_type=ClockType.perf_counter):
        self._iter_impl = None
        self._cpython__number_of_iterations = 0
        self._cpython__last_tracked_number_of_iterations = 0
        self._cpython__start_time = 0
        self._cpython__end_time = 0
        self._cpython__relevant_stop_time = 0
        self._cpython__last_run_was_made = False
        self._cpython__clock_type = clock_type
        self._cpython__clock = None
        self._init_clock()
        self._cpython__run_time = <clock_t>round(run_time * <float>CLOCKS_PER_SEC)
        if not self._cpython__run_time:
            raise TimeLimitIsTooSmall(1.0 / <float>CLOCKS_PER_SEC)

    def __call__(self, *args, **kwargs):
        return self._cpython__last_run_was_made
    
    def _init_clock(self):
        if ClockType.fake == self._cpython__clock_type:
            self._cpython__clock = _fake
        elif ClockType.clock == self._cpython__clock_type:
            self._cpython__clock = _cpython__clock
        elif ClockType.perf_counter == self._cpython__clock_type:
            self._cpython__clock = _perf_counter
        elif ClockType.process_time == self._cpython__clock_type:
            self._cpython__clock = _process_time

    def iter(self):
        raise NotImplemented()

    @property
    def iter_per_time_unit(self):
        raise NotImplemented()

    @property
    def iterations_made(self):
        return self._cpython__last_tracked_number_of_iterations

    @property
    def total_number_of_iterations_made(self):
        return self._cpython__number_of_iterations

    @property
    def time_spent(self):
        return <float>(self._cpython__relevant_stop_time - self._cpython__start_time) / <float>CLOCKS_PER_SEC

    @property
    def total_amount_of_time_spent(self):
        return <float>(self._cpython__end_time - self._cpython__start_time) / <float>CLOCKS_PER_SEC

    @property
    def clock_type(self):
        return self._cpython__clock_type

    # ===============
    @property
    def _clock_type(self):
        return self._cpython__clock_type

    @property
    def _last_tracked_number_of_iterations(self):
        return self._cpython__last_tracked_number_of_iterations

    @property
    def _number_of_iterations(self):
        return self._cpython__number_of_iterations
    
    @property
    def _start_time(self):
        return <float>self._cpython__start_time / <float>CLOCKS_PER_SEC
    
    @property
    def _end_time(self):
        return <float>self._cpython__end_time / <float>CLOCKS_PER_SEC
    
    @property
    def _relevant_stop_time(self):
        return <float>self._cpython__relevant_stop_time / <float>CLOCKS_PER_SEC
    
    @property
    def _last_run_was_made(self):
        return self._cpython__last_run_was_made
    
    def _clock_py(self):
        return <float>self._cpython__clock() / <float>CLOCKS_PER_SEC
    
    @property
    def _clock(self):
        return self._clock_py
    
    @property
    def _run_time(self):
        return <float>self._cpython__run_time / <float>CLOCKS_PER_SEC

    @clock_type.setter
    def clock_type(self, value: ClockType):
        self._cpython__clock_type = value
        self._init_clock()

    @_last_tracked_number_of_iterations.setter
    def _last_tracked_number_of_iterations(self, value):
        self._cpython__last_tracked_number_of_iterations = value

    @_number_of_iterations.setter
    def _number_of_iterations(self, value):
        self._cpython__number_of_iterations = value

    @_clock_type.setter
    def _clock_type(self, value: ClockType):
        self._cpython__clock_type = value
    
    @_start_time.setter
    def _start_time(self, value):
        self._cpython__start_time = <clock_t>round(value * <float>CLOCKS_PER_SEC)
    
    @_end_time.setter
    def _end_time(self, value):
        self._cpython__end_time = <clock_t>round(value * <float>CLOCKS_PER_SEC)
    
    @_relevant_stop_time.setter
    def _relevant_stop_time(self, value):
        self._cpython__relevant_stop_time = <clock_t>round(value * <float>CLOCKS_PER_SEC)
    
    @_last_run_was_made.setter
    def _last_run_was_made(self, value):
        self._cpython__last_run_was_made = value
    
    @_run_time.setter
    def _run_time(self, value):
        self._cpython__run_time = <clock_t>round(value * <float>CLOCKS_PER_SEC)


cdef class Tracer(BaseTracer):
    """
    Main tracer.
    Its task is to find out the speed of code execution, and to stop the counting at about the specified time.

    Example of use:

        tr = Tracer(10.0)
        while tr.iter():
            i = '456'
            k = int('1243' + i)

        print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    """

    cdef:
        clock_t _cpython__half_of_the_run_time
        clock_t _cpython__relevant_start_time
        unsigned int _cpython__relevant_number_of_iterations_at_start
        unsigned int _cpython__relevant_number_of_iterations_at_end
        unsigned int _cpython__next_test_stop_on
        object _cpython__testing_worker

    def __init__(self, float run_time, object clock_type=ClockType.perf_counter):
        super().__init__(run_time, clock_type)
        self._iter_impl = self._first_run
        self._cpython__testing_worker = self._test_runs
        self._cpython__relevant_start_time = 0
        self._cpython__relevant_number_of_iterations_at_start = 0
        self._cpython__relevant_number_of_iterations_at_end = 0
        self._cpython__next_test_stop_on = 1
        self._cpython__half_of_the_run_time = <clock_t>round(<float>self._cpython__run_time / 2.0)
        if not self._cpython__half_of_the_run_time:
            raise TimeLimitIsTooSmall(2.0 / <float>CLOCKS_PER_SEC)

    def iter(self):
        return self._iter_impl()

    @property
    def iter_per_time_unit(self):
        cdef:
            float divider

        if self._cpython__last_run_was_made:
            divider = <float>(self._cpython__relevant_stop_time - self._cpython__relevant_start_time) / <float>CLOCKS_PER_SEC
            if 0.0 != divider:
                return <float>(self._cpython__relevant_number_of_iterations_at_end - self._cpython__relevant_number_of_iterations_at_start) \
                       / divider
            return 0.0
        else:
            divider = <float>(self._cpython__end_time - self._cpython__start_time) / <float>CLOCKS_PER_SEC
            if 0.0 != divider:
                return <float>self._cpython__last_tracked_number_of_iterations / divider
            return 0.0

    def _first_run(self):
        self._cpython__number_of_iterations += 1
        self._cpython__relevant_start_time = self._cpython__start_time = self._cpython__clock()
        self._iter_impl = self._subsequent_runs
        return self._test_sub_runs(self._cpython__start_time)

    def _subsequent_runs(self):
        self._cpython__number_of_iterations += 1
        if self._cpython__number_of_iterations >= self._cpython__next_test_stop_on:
            return self._cpython__testing_worker()
        return True

    def _test_runs(self):
        return self._test_sub_runs(self._cpython__clock())

    def _test_sub_runs(self, clock_t current_time):
        cdef:
            clock_t delta_time
            clock_t time_left
            float iterations_per_second
            float iterations_left

        self._cpython__relevant_stop_time = self._cpython__end_time = current_time
        self._cpython__last_tracked_number_of_iterations = self._cpython__number_of_iterations
        delta_time = current_time - self._cpython__start_time
        if delta_time >= self._cpython__half_of_the_run_time:
            time_left = self._cpython__run_time - delta_time
            # No need to:
            # if time_left < 0:
            #     time_left = 0
            iterations_per_second = <float>self._cpython__number_of_iterations / <float>delta_time
            iterations_left = iterations_per_second * time_left
            if iterations_left > 0.0:
                self._cpython__next_test_stop_on = self._cpython__number_of_iterations + round(iterations_left)
                self._cpython__testing_worker = self._last_run
                self._cpython__relevant_start_time = current_time
                self._cpython__relevant_number_of_iterations_at_start = self._cpython__number_of_iterations
                return True
            else:
                return self._last_sub_run(current_time)
        self._cpython__next_test_stop_on *= 2
        return True

    def _last_run(self):
        return self._last_sub_run(self._cpython__clock())

    def _last_sub_run(self, clock_t current_time):
        self._cpython__relevant_stop_time = self._cpython__end_time = current_time
        self._cpython__last_tracked_number_of_iterations = self._cpython__number_of_iterations
        self._cpython__relevant_number_of_iterations_at_end = self._cpython__number_of_iterations
        self._cpython__testing_worker = self._after_last_runs
        self._cpython__last_run_was_made = 1
        return False

    def _after_last_runs(self):
        self._cpython__end_time = self._cpython__clock()
        return False
    
    @property
    def _testing_worker(self):
        return self._cpython__testing_worker
    
    @property
    def _relevant_start_time(self):
        return <float>self._cpython__relevant_start_time / <float>CLOCKS_PER_SEC
    
    @property
    def _relevant_number_of_iterations_at_start(self):
        return self._cpython__relevant_number_of_iterations_at_start
    
    @property
    def _relevant_number_of_iterations_at_end(self):
        return self._cpython__relevant_number_of_iterations_at_end
    
    @property
    def _next_test_stop_on(self):
        return self._cpython__next_test_stop_on
    
    @property
    def _half_of_the_run_time(self):
        return <float>self._cpython__half_of_the_run_time / <float>CLOCKS_PER_SEC
    
    @_testing_worker.setter
    def _testing_worker(self, value):
        self._cpython__testing_worker = value
    
    @_relevant_start_time.setter
    def _relevant_start_time(self, value):
        self._cpython__relevant_start_time = <clock_t>round(value * <float>CLOCKS_PER_SEC)
    
    @_relevant_number_of_iterations_at_start.setter
    def _relevant_number_of_iterations_at_start(self, value):
        self._cpython__relevant_number_of_iterations_at_start = value
    
    @_relevant_number_of_iterations_at_end.setter
    def _relevant_number_of_iterations_at_end(self, value):
        self._cpython__relevant_number_of_iterations_at_end = value
    
    @_next_test_stop_on.setter
    def _next_test_stop_on(self, value):
        self._cpython__next_test_stop_on = value
    
    @_half_of_the_run_time.setter
    def _half_of_the_run_time(self, value):
        self._cpython__half_of_the_run_time = <clock_t>round(value * <float>CLOCKS_PER_SEC)


cdef class GreedyTracer(BaseTracer):
    """
    Greedy Main tracer.
    Its task is to find out the speed of code execution, and to stop the counting at about the specified time.
    The difference is that he checks time every single iteration.

    Example of use is the same as for the Tracer()

    """

    def __init__(self, run_time: float, clock_type=ClockType.perf_counter):
        super().__init__(run_time, clock_type)
        self._iter_impl = self._first_run

    def iter(self):
        return self._iter_impl()

    def _first_run(self):
        self._cpython__start_time = self._cpython__clock()
        self._iter_impl = self._subsequent_runs
        return self._subsequent_runs()

    def _subsequent_runs(self):
        self._cpython__relevant_stop_time = self._cpython__end_time = self._cpython__clock()
        self._cpython__number_of_iterations += 1
        self._cpython__last_tracked_number_of_iterations = self._cpython__number_of_iterations

        if (self._cpython__relevant_stop_time - self._cpython__start_time) < self._cpython__run_time:
            return True
        else:
            self._cpython__last_run_was_made = True
            self._iter_impl = self._after_last_runs
            return False

    def _after_last_runs(self):
        self._cpython__number_of_iterations += 1
        self._cpython__end_time = self._cpython__clock()
        return False

    @property
    def iter_per_time_unit(self):
        cdef:
            float divider

        divider = <float>(self._cpython__relevant_stop_time - self._cpython__start_time) / <float>CLOCKS_PER_SEC
        if 0 != divider:
            return <float>self._cpython__last_tracked_number_of_iterations / divider
        return 0


cdef class TracerCounter(BaseTracer):
    """
    Counting tracer. Pseudo-tracer.
    Its don't have an overhead of periodic calling time() function.
    Its task is to count down within a given time, using the speed information already counted by the real tracer
    (by the Tracer class).

    Example of use:

        trc = TracerCounter(10000, 10.0)
        while trc.iter():
            i = '456'
            k = int('1243' + i)

        print('{} iter/s; {} seconds; {} iters'.format(trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

    or:
        def made_tests()->Tracer:
            tr = Tracer(0.1)  # Run for about 0.1 of second.
            while tr.iter():
                some_my_code()
            return tr

        def run(run_time: float, tests_result: Tracer):
            trc = TracerCounter(tests_result.iter_per_time_unit, run_time)
            while trc.iter():
                some_my_code()

            print('{} iter/s; {} seconds; {} iterations'.format(trc.iter_per_time_unit, trc.time_spent,
                    trc.iterations_made))

        def main():
            tests_result = made_tests()
            ...
            while True:
                time_to_rur_str = input('Enter run time: ')
                if not time_to_run_str:
                    break
                run(float(time_to_rur_str), tests_result)

    """

    cdef:
        float _cpython__iter_per_time_unit
        unsigned int _cpython__number_of_iterations_needed

    def __init__(self, iter_per_time_unit: float, run_time: float, object clock_type=ClockType.fake):
        super().__init__(run_time, clock_type)
        self._cpython__iter_per_time_unit = iter_per_time_unit
        self._iter_impl = self._first_run
        self._cpython__number_of_iterations_needed = <clock_t>round(self._cpython__iter_per_time_unit * (<float>self._cpython__run_time / <float>CLOCKS_PER_SEC))

    def iter(self):
        return self._iter_impl()

    def _first_run(self):
        self._cpython__start_time = self._cpython__clock()
        self._iter_impl = self._subsequent_runs
        return self._subsequent_runs()

    def _subsequent_runs(self):
        self._cpython__number_of_iterations += 1
        self._cpython__last_tracked_number_of_iterations = self._cpython__number_of_iterations
        if self._cpython__number_of_iterations < self._cpython__number_of_iterations_needed:
            return True
        else:
            self._cpython__relevant_stop_time = self._cpython__end_time = self._cpython__clock()
            self._cpython__last_run_was_made = True
            self._iter_impl = self._after_last_runs
            return False

    def _after_last_runs(self):
        self._cpython__number_of_iterations += 1
        self._cpython__end_time = self._cpython__clock()
        return False

    @property
    def iter_per_time_unit(self):
        cdef:
            float divider

        if self._cpython__last_run_was_made:
            divider = <float>(self._cpython__relevant_stop_time - self._cpython__start_time) / <float>CLOCKS_PER_SEC
            if 0.0 != divider:
                return <float>self._cpython__last_tracked_number_of_iterations / divider
            return 0.0
        else:
            return self._cpython__iter_per_time_unit

    @property
    def _iter_per_time_unit(self):
        return self._cpython__iter_per_time_unit
    
    @property
    def _number_of_iterations_needed(self):
        return self._cpython__number_of_iterations_needed

    @_iter_per_time_unit.setter
    def _iter_per_time_unit(self, value):
        self._cpython__iter_per_time_unit = value
    
    @_number_of_iterations_needed.setter
    def _number_of_iterations_needed(self, value):
        self._cpython__number_of_iterations_needed = value


cdef class TracerIterator:
    """
    An iterator class. It converts any type of given tracer into an iterator.

    As result you have an option to use constructions like this:

        for i in TracerIterator(Tracer(20.0)):
            k = int('1243')

    But keep in mind that in this case there will be a bigger overhead. And there will be less CPU time for the payload.

    """

    cdef object _tracer

    def __init__(self, tracer):
        self._tracer = tracer

    def __iter__(self):
        return self

    def __next__(self):
        if self._tracer.iter():
            return self._tracer.iterations_made
        else:
            raise StopIteration()

    def next(self):
        return self.__next__()

    @property
    def tracer(self):
        return self._tracer



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import *
import ctypes
import ctypes.wintypes as wintypes
from ctypes import windll
import shutil
from cengal.code_flow_control.chained_flow.versions.v_1.chained_flow import ResultType, Chain, \
    ChainValRunner, CriteriaType, chain_reader, \
    ChainInternalResult, ResultExistence

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


_INSTALLING_BUFFER = {
    'copy': shutil.copy,
    'copy2': shutil.copy2,
}


def copy(src_name, dst_name):
    cont = Chain('copy2', None, ResultType(CriteriaType.optional, set()))
    c_ret = ChainValRunner(cont, lambda result: result != 0, judge)
    c_handle = ChainValRunner(cont, lambda result: result.value != INVALID_HANDLE_VALUE_CTYPES, judge)

    c_ret()(ctypes.windll.kernel32.CopyFileW(src_name, dst_name, False))
    c_ret()(ctypes.windll.kernel32.SetFileAttributesW(dst_name, FILE_ATTRIBUTE_NORMAL))
    destination_file_handle = c_handle('open Destination file')(
        create_file(dst_name, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL))
    try:
        creation_time = ctypes.wintypes.FILETIME()
        last_access_time = ctypes.wintypes.FILETIME()
        last_write_time = ctypes.wintypes.FILETIME()
        c_ret()(ctypes.windll.kernel32.SetFileTime(destination_file_handle,
                                                 ctypes.byref(creation_time),
                                                 ctypes.byref(last_access_time),
                                                 ctypes.byref(last_write_time)))
    except:
        raise
    finally:
        c_ret(block_results_criteria=ResultType(CriteriaType.needed, {'open Destination file'}))(
            ctypes.windll.kernel32.CloseHandle(destination_file_handle))

    verdict_exceptions_consolidator(cont, close=True)


def copy2(src_name, dst_name):
    ret = ctypes.windll.kernel32.CopyFileW(src_name, dst_name, False)
    if not ret:
        raise ctypes.WinError()


def create_file(filename, access, mode, security_attributes, creation, flags, template_file_handle):
    """See: CreateFile function
    http://msdn.microsoft.com/en-us/library/windows/desktop/aa363858(v=vs.85).aspx
    """
    create_file_spec = windll.kernel32.CreateFileW
    create_file_spec.argtypes = [
        wintypes.LPWSTR,  # _In_ - LPCTSTR lpFileName
        wintypes.DWORD,  # _In_ - DWORD dwDesiredAccess
        wintypes.DWORD,  # _In_ - DWORD dwShareMode
        LPSECURITY_ATTRIBUTES,  # _In_opt_ - LPSECURITY_ATTRIBUTES lpSecurityAttributes
        wintypes.DWORD,  # _In_ - DWORD dwCreationDisposition
        wintypes.DWORD,  # _In_ - DWORD dwFlagsAndAttributes
        wintypes.HANDLE]  # _In_opt_ - HANDLE hTemplateFile
    create_file_spec.restype = wintypes.HANDLE

    return wintypes.HANDLE(create_file_spec(filename,
                                         access,
                                         mode,
                                         security_attributes,
                                         creation,
                                         flags,
                                         template_file_handle))


def install_module():
    shutil.copy = copy
    shutil.copy2 = copy2


def uninstall_module():
    shutil.copy = _INSTALLING_BUFFER['copy']
    shutil.copy2 = _INSTALLING_BUFFER['copy2']



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


#!/usr/bin/env python
# coding=utf-8




"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""





import asyncio
from time import perf_counter
# from efficient_streams import *
from cengal.parallel_execution.asyncio.efficient_streams import *
from cengal.hardware.info.cpu import cpu_info
from cengal.io.used_ports import *
from random import random
import marshal
import pickle


asyncio.selector_events._SelectorTransport.max_size = cpu_info().l2_cache_size_per_virtual_core
asyncio.selector_events._DEFAULT_LIMIT = cpu_info().l2_cache_size_per_virtual_core


class PickleEncodableClass:
    def __init__(self) -> None:
        self.name = 'ad;slfakldjgldskfj'
        self.d = [{
            'asdf': 243.0,
            2: [0] * 3000,
            ('asdf', 345): [('erihtserkdhtg', (3, 4, 5))] * 100000,
            ('hjg', 55): [('erihtserkdhtg', (3, 4, 5))] * 100000,
        }] * 100000

st = perf_counter()
dumped_data = pickle.dumps(PickleEncodableClass())
dt = perf_counter() - st
print(f'PickleEncodableClass size: {len(dumped_data)}; time: {dt}')


def randomized_data_size(data_size: int) -> int:
    offset = 0.1
    range = 10.0
    ran = random()
    result = offset + range * ran
    return round(data_size * result)


async def handle_echo(reader, writer):
    data = await reader.read(100)
    message = data.decode()
    addr = writer.get_extra_info('peername')

    print(f"Received {message!r} from {addr!r}")

    print(f"Send: {message!r}")
    # for i in range(100):
    #     writable_data = data * 1000000
    #     # print(f'writable_data len: {len(writable_data)}')
    #     writer.write(writable_data)

    data_chunk_len = int(cpu_info().l2_cache_size_per_virtual_core / len(data))
    pickle_data = pickle.dumps(PickleEncodableClass())
    marsharl_data = marshal.dumps(data * randomized_data_size(data_chunk_len))
    stime = perf_counter()
    dtime = 0
    return_time = 10
    index = 0
    while dtime < return_time:
        writable_data = data * randomized_data_size(data_chunk_len)
        # print(f'writable_data len: {len(writable_data)}')
        index += 1
        if 10 <= index:
            # writer.write(pickle.dumps(PickleEncodableClass()))
            writer.write(pickle_data)
            index = 0
        else:
            # writer.write(marshal.dumps(writable_data))
            writer.write(marsharl_data)

        await writer.drain()
        dtime = perf_counter() - stime

    print("Close the connection")
    writer.close()

async def handle_echo_1(reader, writer):
    reader: TcpStreamReader = reader
    writer: TcpStreamWriter = writer

    data = await reader.read(100)
    message = data.decode()
    addr = writer.get_extra_info('peername')

    print(f"Received {message!r} from {addr!r}")

    print(f"Send: {message!r}")
    # for i in range(100):
    #     writable_data = data * 1000000
    #     # print(f'writable_data len: {len(writable_data)}')
    #     writer.write(writable_data)

    data_chunk_len = int(cpu_info().l2_cache_size_per_virtual_core / len(data))
    stime = perf_counter()
    dtime = 0
    return_time = 10
    index = 0
    while dtime < return_time:
        index += 1
        if 10 <= index:
            writer.owrite(pickle.dumps(PickleEncodableClass()))
            index = 0
        else:
            writer.owrite(marshal.dumps(data * randomized_data_size(data_chunk_len)))

        await writer.partial_drain()
        dtime = perf_counter() - stime
    
    await writer.full_drain()

    print("Close the connection")
    writer.close()


async def handle_echo_2(reader, writer):
    reader: TcpStreamReader = reader
    writer: TcpStreamWriter = writer

    data = await reader.read(100)
    message = data.decode()
    addr = writer.get_extra_info('peername')

    print(f"Received {message!r} from {addr!r}")

    print(f"Send: {message!r}")
    # for i in range(100):
    #     writable_data = data * 1000000
    #     # print(f'writable_data len: {len(writable_data)}')
    #     writer.write(writable_data)

    data_chunk_len = int(cpu_info().l2_cache_size_per_virtual_core / len(data))
    pickle_data = pickle.dumps(PickleEncodableClass())
    marsharl_data = marshal.dumps(data * randomized_data_size(data_chunk_len))
    writer.start_aw()
    stime = perf_counter()
    dtime = 0
    return_time = 10
    index = 0
    while dtime < return_time:
        index += 1
        if 10 <= index:
            # writer.owrite(pickle.dumps(PickleEncodableClass()))
            writer.owrite(pickle_data)
            index = 0
        else:
            # writer.owrite(marshal.dumps(data * randomized_data_size(data_chunk_len)))
            writer.owrite(marsharl_data)

        await writer.aw_drain_enough()
        dtime = perf_counter() - stime
    
    await writer.full_drain()

    print("Close the connection")
    writer.close()


async def main():
    port = purify_ports(used_ports().port(Protocol.tcp, used_ports().tables[Table.user], {PortStatus.no})())
    # server = await asyncio.start_server(
    #     handle_echo, '127.0.0.1', port)
    # server = await TcpStreamManager().start_server(
    #     handle_echo, '127.0.0.1', port)
    # server = await TcpStreamManager().start_server(
    #     handle_echo_1, '127.0.0.1', port)
    server = await TcpStreamManager().start_server(
        handle_echo_2, '127.0.0.1', port)

    addrs = ', '.join(str(sock.getsockname()) for sock in server.sockets)
    print(f'Serving on {addrs}')

    async with server:
        await server.serve_forever()


asyncio.run(main())



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import json
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import gly, CoroPriority

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class RequestCache:
    # TODO: add GC which will clean lists every X seconds (really - every Y hours)

    def __init__(self, itemsQntLimit, timeLimitInSeconds=None, clock_function=None, default_priority: CoroPriority = CoroPriority.normal):
        # timeLimitInSeconds:
        # - number - limit in seconds (remember that the accuracy of the server clock is
        #   approximately 1 second);
        # - 0 (zero) - cache is not used (try_to_get_data_for_request() will return 'None'
        #   on every request);
        # - None - cache is trying to be permanent (time limit is not used at all)
        super().__init__()
        self.default_priority: CoroPriority = default_priority
        self._clock_function = clock_function or time.perf_counter
        self._itemsQntLimit = itemsQntLimit
        self._timeLimitInSeconds = timeLimitInSeconds
        self._requestsAndData = {}   # key - request; data - (data, Server Time of last change). Server Time of last
            #  change must be less or equal to self._itemsQntLimit
        self._requestsHistory = []   # [request0, request1, ..., requestN]
        self.isWasChanged = False

    def put_new_request(self, request, data):
        self._move_request_to_the_end_of_history(request)
        if request in self._requestsAndData:
            dataAndTime = self._requestsAndData[request]
            is_was_changed = False
            if isinstance(data, type(dataAndTime[0])):
                if data != dataAndTime[0]:
                    is_was_changed = True
            else:
                is_was_changed = True

            if is_was_changed:
                self._requestsAndData[request] = (data, self._clock_function())
                self.isWasChanged = True
        else:
            if len(self._requestsAndData) >= self._itemsQntLimit:
                forDeleting = self._requestsHistory[0]
                if forDeleting in self._requestsAndData:
                    will_be_deleted = self._requestsAndData[forDeleting]
                    del self._requestsAndData[forDeleting]
                    return will_be_deleted
                del self._requestsHistory[0]
            self._requestsAndData[request] = (data, self._clock_function())
            self.isWasChanged = True

    def put_new_request_or_renew_it(self, request, data):
        self._move_request_to_the_end_of_history(request)
        if request in self._requestsAndData:
            dataAndTime = self._requestsAndData[request]
            self._requestsAndData[request] = (data, self._clock_function())
            self.isWasChanged = True
        else:
            if len(self._requestsAndData) >= self._itemsQntLimit:
                forDeleting = self._requestsHistory[0]
                if forDeleting in self._requestsAndData:
                    will_be_deleted = self._requestsAndData[forDeleting]
                    del self._requestsAndData[forDeleting]
                    return will_be_deleted
                del self._requestsHistory[0]
            self._requestsAndData[request] = (data, self._clock_function())
            self.isWasChanged = True

    def try_to_get_raw_data_for_request(self, request):
        if request in self._requestsAndData:
            dataAndTime = self._requestsAndData[request]
            return dataAndTime[0]
        else:
            return None

    def try_to_get_raw_data_with_time_for_request(self, request):
        if request in self._requestsAndData:
            dataAndTime = self._requestsAndData[request]
            return dataAndTime
        else:
            return None

    def try_to_get_data_for_request(self, request):
        if request in self._requestsAndData:
            dataAndTime = self._requestsAndData[request]
            lastChangingTime = dataAndTime[1]
            tLimit = self._timeLimitInSeconds
            if (tLimit is not None) and ((self._clock_function() - lastChangingTime) >= tLimit):
                del self._requestsAndData[request]
                if request in self._requestsHistory:
                    self._requestsHistory.remove(request)
                return None
            else:
                self._move_request_to_the_end_of_history(request)
                return dataAndTime[0]
        else:
            return None

    def try_to_get_data_for_request_and_renew_it(self, request):
        if request in self._requestsAndData:
            dataAndTime = self._requestsAndData[request]
            lastChangingTime = dataAndTime[1]
            tLimit = self._timeLimitInSeconds
            if (tLimit is not None) and ((self._clock_function() - lastChangingTime) >= tLimit):
                del self._requestsAndData[request]
                if request in self._requestsHistory:
                    self._requestsHistory.remove(request)
                return None
            else:
                self._move_request_to_the_end_of_history(request)
                self._requestsAndData[request] = (dataAndTime[0], self._clock_function())
                self.isWasChanged = True
                return dataAndTime[0]
        else:
            return None

    def try_to_remove_request(self, request):
        if request in self._requestsAndData:
            del self._requestsAndData[request]
        if request in self._requestsHistory:
            self._requestsHistory.remove(request)

    def _move_request_to_the_end_of_history(self, request):
        if request in self._requestsAndData:
            if request in self._requestsHistory:
                self._requestsHistory.remove(request)
        self._requestsHistory.append(request)

    def update(self, anotherRequestCache):
        if type(anotherRequestCache) == RequestCache:
            self._itemsQntLimit += anotherRequestCache._itemsQntLimit
            self._requestsAndData.update(anotherRequestCache._requestsAndData)
            self.isWasChanged = True
            # Do not do this:
            # self._requestsHistory += anotherRequestCache._requestsHistory

    def clear(self):
        self._requestsAndData = type(self._requestsAndData)()
        # self._requestsHistory.clear()  # Python 3.3+ only so can't be used under PyPy yet.
        self._requestsHistory = type(self._requestsHistory)()
        self.isWasChanged = True

    def get_state(self):
        reqAndDat = []
        for item in self._requestsAndData.items():
            reqAndDat.append(item)
        data = (self._itemsQntLimit
                , self._timeLimitInSeconds
                , reqAndDat
                , self._requestsHistory)
        return json.dumps(data)

    def set_state(self, state):
        data = json.loads(state)
        self._itemsQntLimit = data[0]
        self._timeLimitInSeconds = data[1]
        for item in data[2]:
            self._requestsAndData[item[0]] = item[1]
        self._requestsHistory = data[3]
        self.isWasChanged = True



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.code_flow_control.multiinterface_essence import *


Kilogram = float


class Weight:
    def __init__(self, weight: Kilogram):
        self.weight: Kilogram = weight


class Point:
    def __init__(self, x: float, y: float):
        self.x: float = x
        self.y: float = y


class MaterialPoint:
    def __init__(self, material_point: Point):
        self.material_point = material_point


class Quadrangle:
    def __init__(self, tl: Point, tr: Point, bl: Point, br: Point):
        self.tl: Point = tl
        self.tr: Point = tr
        self.bl: Point = bl
        self.br: Point = br


class StretchingQuadrangleFrameModel(EssenceModel, MaterialPoint, Quadrangle, Weight):
    def __init__(self):
        super(StretchingQuadrangleFrameModel, self).__init__()
        Quadrangle.__init__(self, Point(0, 0), Point(0, 0), Point(0, 0), Point(0, 0))
        MaterialPoint.__init__(self, Point(0, 0))
        Weight.__init__(self, 0)


class Vector:
    def __init__(self, start: Point, end: Point):
        self.start = start
        self.end = end


class Flying:
    def __init__(self, flying_vector: Vector):
        self.flying_vector = flying_vector


class FlyingStretchingQuadrangleFrameModel(StretchingQuadrangleFrameModel, Flying):
    def __init__(self):
        super(FlyingStretchingQuadrangleFrameModel, self).__init__()
        Flying.__init__(self, Vector(Point(0, 0), Point(0, 0)))



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import platform
if 'PyPy' != platform.python_implementation():
    import requests
import binascii
import os, os.path
import pickle
import datetime
# try to import C parser then fallback in pure python parser.
try:
    from http_parser.parser import HttpParser
except ImportError:
    from http_parser.pyparser import HttpParser

from cengal.modules_management.alternative_import import alt_import

with alt_import('lzma') as lzma:
    if lzma is None:
        import lzmaffi.compat
        lzmaffi.compat.register()
        import lzma


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def remove_percent_encoding_from_the_URI(string, plus=True):
    try:
        string = str(string)
        if plus:
            string = string.replace('+', ' ')
        string = string.encode(encoding='utf-8')
        percentTypes = ((b'%u', 6, (False, None)), (b'%', 3, (True, tuple(range(32)) + (34, 42, 58, 60, 62, 63, 124, 127))))
        # (prefix, full size, (allowed type or not at all, list of disallowed characters like backspace etc.))
        for percType in percentTypes:
            isDone = False
            while not isDone:
                index = string.find(percType[0])
                if index > -1:
                    ind2 = index+percType[1]
                    if ind2 > len(string):
                        ind2 = len(string)
                    hexString = string[index:ind2]
                    hexString = hexString[len(percType[0]):]
                    decodedString = b''
                    if percType[2][0]:
                        try:
                            decodedStringBuff = binascii.unhexlify(hexString)
                            if int.from_bytes(decodedStringBuff, byteorder='little') not in percType[2][1]:
                                decodedString = decodedStringBuff
                        except binascii.Error:
                            pass
                    string = string[:index] + decodedString + string[ind2:]
                else:
                    isDone = True
            if percType[2][0]:
                if (len(string) > 0) and (len(set(string).intersection(set(percType[2][1]))) > 0):
                    index = 0
                    isDone = False
                    while not isDone:
                        if string[index] in percType[2][1]:
                            string = string[:index] + string[index+1:]
                        else:
                            index += 1
                        if index >= len(string):
                            isDone = True
        string = string.decode(encoding='utf-8')
    except UnicodeDecodeError:
        string = None
    except UnicodeEncodeError:
        string = None
    return string


def get_standard_folder_separator():
    return '/'


def unify_folder_separators(string):
    string = str(string)
    string = string.replace('\\', get_standard_folder_separator())
    return string


def remove_forbidden_file_names_from_the_URI(string):
    # replace forbidden file names with slash ('/'). Also will unify folder separators (replace '\' with '/')
    forbiddenFileNames = {'con', 'prn', 'aux', 'nul', 'com1', 'com2', 'com3', 'com4', 'com5', 'com6', 'com7', 'com8'
        , 'com9', 'lpt1', 'lpt2', 'lpt3', 'lpt4', 'lpt5', 'lpt6', 'lpt7', 'lpt8', 'lpt9'}
    string = unify_folder_separators(str(string))
    strBuffer = string
    string = string.lower()
    for forbFile in forbiddenFileNames:
        string = string.replace(get_standard_folder_separator() + forbFile + get_standard_folder_separator(), '/')
        string = string.replace(get_standard_folder_separator() + forbFile + '.', '/')
    isStrWasChanged = False
    if strBuffer.lower() != string:
        isStrWasChanged = True
    else:
        string = strBuffer
    result = (string, isStrWasChanged)
    return result


def remove_path_to_the_parent_of_the_current_directory(string):
    string = unify_folder_separators(str(string))
    string = string.replace('../', '/')
    string = string.replace('/..', '/')
    return string


def is_path_is_trying_to_leave_site_sandbox(string):
    itemsList = string.split(get_standard_folder_separator())
    while '' in itemsList:
        itemsList.remove('')
    counter = 0
    for item in itemsList:
        if item == '..':
            counter -= 1
        elif item == '.':
            pass
        else:
            counter += 1
        if counter < 0:
            return True
    return False


def is_path_is_not_safe(string):
    # return tuple (decoded string, is_path_is_not_safe)
    result = tuple()
    string = remove_percent_encoding_from_the_URI(string)
    if string is None:
        result = (string, True)
        return result
    string = remove_forbidden_file_names_from_the_URI(string)
    if string[1]:
        result = (string[0], True)
        return result
    is_not_safe = is_path_is_trying_to_leave_site_sandbox(string[0])
    result = (string[0], is_not_safe)
    return result


def web_server__is_redirection_to_the_main_domain_needed(httpParser: HttpParser, prefix=None):
    functionResult = False

    host = httpParser.get_headers()['Host']
    domain = host
    if ':' in host:
        host = host.split(':')
        while '' in host:
            host.remove('')
        domain = host[0]
    if '.' in domain:
        domain = domain.split('.')
        if domain[0] == '':
            del domain[0]
            functionResult = True
        if prefix is not None:
            if domain[0] != prefix:
                domain.insert(0, prefix)
                functionResult = True
        lastDLen = len(domain) - 1
        if domain[lastDLen] == '':
            del domain[lastDLen]
            functionResult = True
        domain = '.'.join(domain)
    if host.__class__ is list:
        if len(host) > 1:
            host[0] = domain
            domain = ':'.join(host)

    functionResult = (functionResult, domain)
    return functionResult


def load_cache_from_file(fileName, originalCache):
    fileName = unify_folder_separators(fileName)
    cache = originalCache
    if os.path.exists(fileName):
        if os.path.isfile(fileName):
            cacheDataFile = None
            dataBuff = originalCache
            try:
                # cacheDataFile = open(fileName, 'rb')
                cacheDataFile = lzma.open(fileName, 'rb')
                cache = pickle.load(cacheDataFile)
                dataBuff.update(cache)
                cache = dataBuff
            except(EnvironmentError, pickle.PicklingError) as err:
                cache = dataBuff
            finally:
                if cacheDataFile is not None:
                    cacheDataFile.close()
    return cache


def save_cache_to_file(fileName, cache):
    fileName = unify_folder_separators(fileName)
    if cache.isWasChanged:
        isRenamedToBak = False
        if os.path.exists(fileName):
            try:
                if os.path.exists(fileName+'.bak'):
                    try:
                        os.remove(fileName+'.bak')
                    except IOError:
                        pass
                os.rename(fileName, fileName+'.bak')
                isRenamedToBak = True
            except IOError:
                pass

        isDumped = False
        cacheDataFile = None
        try:
            # cacheDataFile = open(fileName, 'wb')
            cacheDataFile = lzma.open(fileName, 'wb', format=lzma.FORMAT_XZ, check=lzma.CHECK_CRC64, preset=1)
            pickle.dump(cache, cacheDataFile)
            cache.isWasChanged = False
            isDumped = True
        except(EnvironmentError, pickle.PicklingError) as err:
            pass
        finally:
            if cacheDataFile is not None:
                cacheDataFile.close()

        if isRenamedToBak and isDumped:
            try:
                os.remove(fileName+'.bak')
            except IOError:
                pass


class CachedRequestError(Exception): pass


def make_cached_request__get(cache, request, useOnlyCachedResults=False):
    result = cache.try_to_get_data_for_request(request)
    if (result is None) or (not result.ok):
        if useOnlyCachedResults:
            raise CachedRequestError()
        result = requests.get(unify_folder_separators(request))
        result = SerializableHttpResponseFromRequests(result)
        cache.put_new_request(request, result)

    return result


def make_cached_request__universal(cache, request, workerFunction, useOnlyCachedResults=False, **workerFunctionParams):
    result = cache.try_to_get_data_for_request(request)
    if result is None:
        if useOnlyCachedResults:
            raise CachedRequestError()
        result = workerFunction(**workerFunctionParams)
        cache.put_new_request(request, result)

    return result


def make_a_copy_of_cached_request_with_another_keyname(cache, current_name, new_name, make_first_copy_only=False):
    is_copy_avaliable = True
    if make_first_copy_only:
        test = cache.try_to_get_data_for_request(current_name)
        if test is not None:
            is_copy_avaliable = False
    if is_copy_avaliable:
        result = cache.try_to_get_data_for_request(current_name)
        if result is not None:
            cache.put_new_request(new_name, result)


class SerializableHttpResponseFromRequests:
    def __init__(self, request, isBinaryFile=False):
        super(SerializableHttpResponseFromRequests, self).__init__()
        self.isBinaryFile = isBinaryFile
        self.url = request.url
        self.ok = request.ok
        self.reason = request.reason
        self.status_code = request.status_code
        self.headers = request.headers
        self.encoding = request.encoding

        if 'content' in dir(request):
            self.content = request.content
        else:
            self.content = None

        # if not isBinaryFile:
        #     if 'apparent_encoding' in dir(request):
        #         self.apparent_encoding = request.apparent_encoding
        #     else:
        #         self.apparent_encoding = None
        # else:
        #     self.apparent_encoding = None
        self.apparent_encoding = None

        # if (self.encoding is not None) or (self.apparent_encoding is not None):
        #     self.text = request.text
        # else:
        #     self.text = None
        self.text = None

    def getResultEncoding(self):
        if self.encoding is not None:
            return self.encoding
        elif self.apparent_encoding is not None:
            return self.apparent_encoding
        else:
            return 'utf-8'

    def getBytesContent(self):
        if self.content is not None:
            return self.content
        elif self.text is not None:
            bData = bytes(self.text, self.getResultEncoding())
            return bData
        else:
            print("Can't get data from the request object on url ", self.url)
            raise Exception


def saveRequestedFileToFS(folderName, requestResult):
    fileName = folderName + os.path.basename(requestResult.url)
    fileName = unify_folder_separators(fileName)
    with open(fileName, 'wb') as file:
        if requestResult.content is not None:
            file.write(requestResult.content)
        elif requestResult.text is not None:
            bData = bytes(requestResult.text, requestResult.getResultEncoding())
            file.write(bData)
        else:
            ex_text = "Can't save object from {} to the fie. Can't get data to save".format(requestResult.url)
            print(ex_text)
            raise Exception(ex_text)


def getFileModificationDate(fileName):
    fileName = unify_folder_separators(fileName)
    t = os.path.getmtime(fileName)
    return datetime.datetime.fromtimestamp(t)





#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import unittest
from cengal.unittest.behavior_stabilizer import UnittestTestCaseState, UnittestTestCaseWithState
from cengal.math.numbers import ae
from cengal.time_management.run_time import RT

from cengal.parallel_execution.coroutines.coro_tools.run_in_loop import run_in_loop
from cengal.parallel_execution.coroutines.coro_scheduler import Interface, CoroWrapperBase, CoroID
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
from cengal.parallel_execution.coroutines.coro_standard_services.cpu_tick_count_per_second import CpuTickCountPerSecond
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro
from cengal.parallel_execution.coroutines.coro_standard_services.kill_coro import KillCoro
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_loop import ShutdownLoop
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_on_keyboard_interrupt import ShutdownOnKeyboardInterrupt
from cengal.time_management.cpu_clock_cycles import perf_counter
from time import perf_counter
from typing import Tuple


async def sub_background_coro(i: Interface, period: float):
    while True:
        await i(Sleep, period)
        print(f'{perf_counter()}')


def background_coro(i: Interface, period: float):
    i(PutCoro, sub_background_coro, period / 2)
    while True:
        i(Sleep, period)
        print(f'{perf_counter()} || {i(CpuTickCountPerSecond)}')


async def await_for_cpu_tick_count(i: Interface) -> Tuple:
    cpu_clock_cycles, last_ticks_per_second, val_99, val_95, val_68, max_deviation, min_deviation = await i(CpuTickCountPerSecond)
    while (last_ticks_per_second is None) or (val_99 < 1) or (val_95 < 1) or (val_68 < 1):
        cpu_clock_cycles, last_ticks_per_second, val_99, val_95, val_68, max_deviation, min_deviation = await i(CpuTickCountPerSecond)
    
    return cpu_clock_cycles, last_ticks_per_second, val_99, val_95, val_68, max_deviation, min_deviation


async def amy_coro(i: Interface, a, b) -> int:
    try:
        await i(ShutdownOnKeyboardInterrupt)
        print('Started', perf_counter())
        print(await await_for_cpu_tick_count(i))
        coro_id: CoroID = await i(PutCoro, background_coro, 0.1)
        await i(Sleep, 0.5)
        await i(KillCoro, coro_id, tree=True)
        print(await i(CpuTickCountPerSecond))
        return a + b
    finally:
        print('Ended', perf_counter())


class TestCaseForRunInLoop(UnittestTestCaseWithState):
    @classmethod
    def setUpClass(cls):
        cls.state = UnittestTestCaseState()

    @classmethod
    def tearDownClass(cls):
        cls.state.unregister()
    
    def test_amain(self):
        with RT() as rt:
            result = self.state.register_outro(run_in_loop)(amy_coro, 2, 5)
        
        self.assertEqual(result, 7, 'Result is wrong')
        self.assertTrue(ae(rt.rt, 0.5, 0.5), f'Run time is wrong: {rt.rt}')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


#!/usr/bin/env python
# coding=utf-8




"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""





import asyncio
# from efficient_streams import *
from cengal.parallel_execution.asyncio.efficient_streams import *
from time import perf_counter
from cengal.hardware.info.cpu import cpu_info
from cengal.io.used_ports import *


# asyncio.selector_events._SelectorTransport.max_size = cpu_info().l2_cache_size_per_virtual_core
# asyncio.selector_events._DEFAULT_LIMIT = cpu_info().l2_cache_size_per_virtual_core


print('STARTED.')


async def read_with_counter(reader):
    if reader._exception is not None:
        raise reader._exception

    # This used to just loop creating a new waiter hoping to
    # collect everything in self._buffer, but that would
    # deadlock if the subprocess sends more than self.limit
    # bytes.  So just call self.read(self._limit) until EOF.
    blocks = []
    counter = 0
    while True:
        block = await reader.read(reader._limit)
        counter += 1
        if not block:
            break
        blocks.append(block)
    return b''.join(blocks), counter


async def tcp_echo_client(message):
    port = purify_ports(used_ports().port(Protocol.tcp, used_ports().tables[Table.user], {PortStatus.no})())
    # reader, writer = await asyncio.open_connection(
    #     '127.0.0.1', port)
    reader, writer = await TcpStreamManager().open_connection(
        '127.0.0.1', port)

    print(f'Send: {message!r}')
    writer.write(message.encode())
    await writer.drain()

    # await asyncio.sleep(10)
    start_time = perf_counter()
    data_len_counter = 0
    io_counter = 0
    data = None
    while data is None or data:
        # data = await reader.read(100 * 1024**2)
        # data, counter = await reader.read()
        # data, counter = await read_with_counter(reader)

        # data, counter = await reader.read_with_counter()
        # io_counter += counter

        data = await reader.read_nearly_max()
        io_counter += 1
        data_len_counter += len(data)
        # print(f'Received: {data.decode()!r}')

    print(f'io_counter: {io_counter}')
    print(f'Bytes received: {data_len_counter}')
    end_time = perf_counter()
    diff_time = end_time - start_time
    if 0 != diff_time:
        speed = data_len_counter / diff_time
        print(f'Bytes per second: {speed}')

    print('Close the connection')
    writer.close()

asyncio.run(tcp_echo_client('Hello World!'))


#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from cengal.code_flow_control.smart_values.versions.v_0 import ResultExistence
from cengal.system import PYTHON_VERSION_INT
import sys

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class DumbProfiler:
    def __call__(self, function):
        return function

    def print_stats(self):
        return

    def dump_stats(self):
        return


try:
    from line_profiler import LineProfiler
except ImportError:
    LineProfiler = DumbProfiler


def set_profiler(allow_profiling=True):
    prof = None
    if allow_profiling:
        prof = LineProfiler()
    else:
        prof = DumbProfiler()
    if sys.version_info >= (3,):
        import builtins
    else:
        import __builtin__ as builtins
    if 'profile' not in builtins.__dict__:
        builtins.__dict__['profile'] = prof
    else:
        if type(builtins.__dict__['profile']) != type(prof):
            raise Exception('Profiler settings are desynced between modules!')


@contextmanager
def profiler_result(profiler, print_result=False, output_file: ResultExistence=None):
    output_file = output_file or ResultExistence(None, None)

    try:
        yield profiler
    except:
        raise
    finally:
        if print_result:
            profiler.print_stats()
        if output_file.existence:
            if output_file.result:
                profiler.dump_stats(output_file.result)
            else:
                # TODO: сделать автоматическое имя из имени прогоняемого файла и случайного GUID
                raise NotImplemented()


